from __future__ import annotations

from datetime import timedelta
from collections import defaultdict
import json
from logging import getLogger
from statistics import mean
from typing import Dict, List, Tuple, Optional, Iterable, Iterator, Union, TYPE_CHECKING
from pathlib import Path

from IPython.display import display, Markdown
from markdown import markdown as to_html
from markdown_strings import esc_format as escape, link, header, bold, italics
from more_itertools import zip_equal
from humanize import naturaldelta
import regex

from ..document import Document, Documents, Sentence
from ..entity import Entity
from ..config import CONFIG as _CONFIG

if TYPE_CHECKING:  # avoid circular dependency
    from ..index import Results, Result, Index


LOGGER = getLogger(__name__)


Duration = timedelta
IndexName = str

SerializedResults = List[List]
EntityRepr = str

AllResults = Tuple[Dict[Entity, 'Results'], Duration, IndexName]
AllSerializedResults = List

NerTags = str


class TaggedSentence:
    CONFIG = _CONFIG['search.TaggedSentence']
    ROOT_PATH = Path(CONFIG['root_path'])

    def __init__(self, sentence: Sentence, ner_tags: NerTags):
        sentence = regex.sub(r'\n+', ' ', sentence)
        assert '\n' not in sentence
        assert '\n' not in ner_tags

        sentence_tuple = tuple(sentence.split())
        ner_tags_tuple = tuple(ner_tags.split())
        if len(sentence_tuple) != len(ner_tags_tuple):
            raise ValueError('Incompatible sentence and NER tags')
        self.sentence_tuple = sentence_tuple
        self.ner_tags_tuple = ner_tags_tuple

    @property
    def sentence(self) -> Sentence:
        sentence = ' '.join(self.sentence_tuple)
        return sentence

    @property
    def ner_tags(self) -> NerTags:
        ner_tags = ' '.join(self.ner_tags_tuple)
        return ner_tags

    @classmethod
    def save(cls, basename: str, tagged_sentences: Iterable['TaggedSentence']) -> None:
        sentences_filename = cls._get_sentences_filename(basename)
        ner_tags_filename = cls._get_ner_tags_filename(basename)
        with sentences_filename.open('wt', encoding='utf8') as sf, \
                ner_tags_filename.open('wt', encoding='utf8') as ntf:
            for tagged_sentence in tagged_sentences:
                print(tagged_sentence.sentence, file=sf)
                print(tagged_sentence.ner_tags, file=ntf)
        LOGGER.debug(f'Saved {sentences_filename} and {ner_tags_filename}')

    @classmethod
    def load(cls, basename: str) -> Iterable['TaggedSentence']:
        sentences_filename = cls._get_sentences_filename(basename)
        ner_tags_filename = cls._get_ner_tags_filename(basename)
        with sentences_filename.open('rt', encoding='utf8') as sf, \
                ner_tags_filename.open('rt', encoding='utf8') as ntf:
            for sentence, ner_tags in zip_equal(sf, ntf):
                sentence = sentence.rstrip('\r\n')
                ner_tags = ner_tags.rstrip('\r\n')
                tagged_sentence = TaggedSentence(sentence, ner_tags)
                yield tagged_sentence
        LOGGER.debug(f'Loaded {sentences_filename} and {ner_tags_filename}')

    @staticmethod
    def deduplicate(tagged_sentences: Iterable['TaggedSentence']) -> Iterable['TaggedSentence']:
        all_ner_tag_lists = defaultdict(lambda: list())
        for tagged_sentence in tagged_sentences:
            all_ner_tag_lists[tagged_sentence.sentence].append(tagged_sentence.ner_tags)
        for sentence, ner_tags_list in all_ner_tag_lists.items():
            combined_ner_tags = None
            for ner_tags in ner_tags_list:
                ner_tags = ner_tags.split()
                if combined_ner_tags is None:
                    combined_ner_tags = ner_tags
                else:
                    assert len(ner_tags) == len(combined_ner_tags)
                    for index, ner_tag in enumerate(ner_tags):
                        if ner_tag != 'O':
                            combined_ner_tags[index] = ner_tag
            assert combined_ner_tags is not None
            tagged_sentence = TaggedSentence(sentence, ' '.join(combined_ner_tags))
            yield tagged_sentence

    @classmethod
    def _get_sentences_filename(cls, basename: str) -> Path:
        filename = (cls.ROOT_PATH / basename).with_suffix('.sentences.txt')
        return filename

    @classmethod
    def _get_ner_tags_filename(cls, basename: str) -> Path:
        filename = (cls.ROOT_PATH / basename).with_suffix('.ner_tags.txt')
        return filename

    def __str__(self) -> str:
        return self.sentence

    def __repr__(self) -> str:
        return str(self)

    def _ipython_display_(self) -> None:
        display(self.markdown)

    @property
    def markdown(self) -> Markdown:
        formatted_sentence_list = []
        for word, ner_tag in zip(self.sentence_tuple, self.ner_tags_tuple):
            if ner_tag == 'O':
                formatted_word = word
            elif ner_tag == 'PER':
                formatted_word = bold(escape(word))
            elif ner_tag == 'LOC':
                formatted_word = italics(escape(word))
            else:
                raise ValueError(f'Unknown tag "{ner_tag}"')
            formatted_sentence_list.append(formatted_word)
        formatted_sentence = ' '.join(formatted_sentence_list)
        return Markdown(formatted_sentence)


def serialize(all_results: AllResults) -> AllSerializedResults:
    results, duration, index_name = all_results
    serialized_results = dict()
    for entity, result in results.items():
        entity_repr = repr(entity)
        serialized_results[entity_repr] = []
        for (document, (start, end)), similarity in result:
            test_snippet = document[start:end]
            serialized_result = [document.filename, start, end, test_snippet, similarity]
            serialized_results[entity_repr].append(serialized_result)
    serialized_duration = duration.total_seconds()
    all_serialized_results = [serialized_results, serialized_duration, index_name]
    return all_serialized_results


def deserialize(all_serialized_results: AllSerializedResults,
                entities: Optional[Iterable[Entity]]) -> AllResults:
    results = defaultdict(lambda: list())
    serialized_results, serialized_duration, index_name = all_serialized_results
    for entity_repr, serialized_result in serialized_results.items():
        entity = Entity.from_repr(entity_repr)
        for filename, start, end, test_snippet, similarity in serialized_result:
            document = Document(filename)
            assert document[start:end] == test_snippet
            result = ((document, (start, end)), similarity)
            results[entity].append(result)
        from ..index import sort_results
        results[entity] = sort_results(results[entity])
    if entities is not None:
        results = {entity: results[entity] for entity in entities}
    duration = timedelta(seconds=serialized_duration)
    all_results = (results, duration, index_name)
    return all_results


class SearchResultList:
    CONFIG = _CONFIG['search.SearchResultList']
    RESULTS_TOPN = CONFIG.getint('number_of_results')
    SENTENCES_TOPN = None if CONFIG['number_of_sentences'] == 'all sentences' else CONFIG.getint('number_of_sentences')
    ROOT_PATH = Path(CONFIG['root_path'])

    def __init__(self, results: Dict[Entity, 'Results'], duration: Duration,
                 index: Union['Index', str]):
        self.results = dict(results)
        self.duration = duration
        self.index_name = str(index)

    def __getitem__(self, entity_or_entities: Union[Iterable[Entity], Entity]) -> Union['Results', 'SearchResultList']:
        if isinstance(entity_or_entities, Entity):
            entity = entity_or_entities
            result = self.results[entity]
            return result
        else:
            entities = entity_or_entities
            results = {entity: self[entity] for entity in entities}
            duration = self.duration * len(entities) / len(self)
            sliced_result_list = SearchResultList(results, duration, self.index_name)
            return sliced_result_list

    def __len__(self) -> int:
        return len(self.results)

    def __iter__(self) -> Iterator[Entity]:
        return iter(self.results)

    def filter_and_map_entities(self, entities: Dict[Entity, Entity]) -> 'SearchResultList':
        updated_results = dict()
        for entity in self:
            if entity not in entities:
                continue
            mapped_entity = entities[entity]
            updated_results[mapped_entity] = self[entity]
        return SearchResultList(updated_results, self.duration, self.index_name)

    def get_tagged_sentences(self, only_relevant: bool, cross_page_boundaries: bool,
                             documents: Optional[Documents]) -> Iterable[TaggedSentence]:
        if only_relevant and cross_page_boundaries:
            assert documents is not None
            documents = {
                basename: document
                for basename, document
                in documents.items()
                if document.is_relevant
            }
        for entity in self:
            for result in self[entity][:self.SENTENCES_TOPN]:
                if TYPE_CHECKING:
                    assert isinstance(result, Result)
                result_head, _ = result
                document, position = result_head
                if only_relevant and not document.is_relevant:
                    continue

                snippet = document.get_sentence(position, cross_page_boundaries, documents)
                if snippet is None:
                    continue

                left_context, match, right_context = snippet
                left_context_tokens = left_context.split()
                match_tokens = match.split()
                right_context_tokens = right_context.split()

                left_context_and_match_tokens = f'{left_context}{match}'.split()
                match_and_right_context_tokens = f'{match}{right_context}'.split()
                assert len(left_context_and_match_tokens) in (
                    len(left_context_tokens) + len(match_tokens),
                    len(left_context_tokens) + len(match_tokens) - 1,
                )
                assert len(match_and_right_context_tokens) in (
                    len(match_tokens) + len(right_context_tokens),
                    len(match_tokens) + len(right_context_tokens) - 1,
                )

                is_left_context_separate = (
                    len(left_context_and_match_tokens) == len(left_context_tokens) + len(match_tokens)
                )
                is_right_context_separate = (
                    len(match_and_right_context_tokens) == len(match_tokens) + len(right_context_tokens)
                )

                ner_tags_list = (
                    (len(left_context_tokens) - 1) * ['O'] +
                    (['O'] if is_left_context_separate and left_context_tokens else []) +
                    len(match_tokens) * [entity.ner_tag] +
                    (len(right_context_tokens) - 1) * ['O'] +
                    (['O'] if is_right_context_separate and right_context_tokens else [])
                )
                ner_tags = ' '.join(ner_tags_list)

                sentence = f'{left_context}{match}{right_context}'

                tagged_sentence = TaggedSentence(sentence, ner_tags)
                yield tagged_sentence

    @property
    def all_results(self) -> AllResults:
        return (self.results, self.duration, self.index_name)

    @classmethod
    def _get_json_filename(cls, basename: str) -> Path:
        filename = (cls.ROOT_PATH / basename).with_suffix('.json')
        return filename

    @classmethod
    def _get_html_filename(cls, basename: str) -> Path:
        filename = (cls.ROOT_PATH / basename).with_suffix('.html')
        return filename

    def save(self, basename: str) -> None:
        self._save_json(basename)
        self._save_html(basename)

    def _save_json(self, basename: str) -> None:
        filename = self.__class__._get_json_filename(basename)
        with filename.open('wt') as f:
            json.dump(serialize(self.all_results), f, indent=4, sort_keys=True)
        LOGGER.debug('Saved {}'.format(filename))

    def _save_html(self, basename: str) -> None:
        filename = self.__class__._get_html_filename(basename)
        with filename.open('wt', encoding='utf8') as f:
            print(self.html, file=f)
        LOGGER.debug('Saved {}'.format(filename))

    @classmethod
    def load(cls, basename: str, entities: Optional[Iterable[Entity]] = None) -> 'SearchResultList':
        filename = cls._get_json_filename(basename)
        with filename.open('rt') as f:
            all_results = deserialize(json.load(f), entities)
        LOGGER.debug('Loaded {}'.format(filename))
        results = SearchResultList(*all_results)
        results._save_html(basename)
        return results

    def __str__(self) -> str:
        lengths = list(map(len, self.results.values()))
        summary = 'Retrieved {} results for {} entities ({:.2f} on average, {} at minimum) in {} using {}.'
        summary = summary.format(sum(lengths), len(self), mean(lengths), min(lengths), naturaldelta(self.duration), self.index_name)
        return summary

    def __repr__(self) -> str:
        return str(self)

    def _ipython_display_(self) -> None:
        display(self.markdown)

    def __radd__(self, other) -> 'SearchResultList':
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def __add__(self, other) -> 'SearchResultList':
        if isinstance(other, SearchResultList):
            assert not (self.results.keys() & other.results.keys())
            results = {**self.results, **other.results}
            duration = self.duration + other.duration
            index_name = 'Combined'
            combined_result_list = SearchResultList(results, duration, index_name)
            return combined_result_list
        raise NotImplementedError

    @property
    def markdown(self) -> Markdown:
        blocks = [str(self)]
        for entity in self:
            result = self[entity][:self.RESULTS_TOPN]
            if TYPE_CHECKING:
                assert isinstance(result, Results)
            top_line = 'Top {} result{} for the entity {}:'
            top_line = top_line.format(len(result), 's' if len(result) > 1 else '', bold(escape(entity)))
            items = []
            for result_number, ((document, position), similarity) in enumerate(result):
                snippet = document.get_markdown_snippet(position).data
                document_name = 'Document {}'.format(escape(repr(document)))
                document_link = link(document_name, document.get_snippet_url(position))
                document_link = bold('{}:'.format(document_link))
                score = italics('(score: {:.2f})'.format(similarity))
                item = '{}. {} {} {}'.format(result_number+1, document_link, snippet, score)
                items.append(item)
            interblock_separator = header('<a href="#{0}" id="{0}">{1}</a>'.format(entity, self.index_name), 1)
            block = '{}\n\n{}\n\n{}'.format(interblock_separator, top_line, '\n'.join(items))
            blocks.append(block)
        document = '\n\n'.join(blocks)
        return Markdown(document)

    @property
    def html(self) -> str:
        title = 'Named entity search results for {}'.format(self.index_name)
        body = to_html(self.markdown.data)
        return '<!DOCTYPE html>\n<meta charset="utf-8">\n<title>{}</title>\n{}'.format(title, body)
